/*
 *   Copyright (c) 2006 by Thanos Kyritsis
 *
 *   This file is part of ffpocket
 *
 *   FFpocket is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   FFpocket is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with FFpocket; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


#include "ffdecode.h"


AVStream *ff_add_audio_stream(AVFormatContext *oc, int codec_id, int sample_rate, int channels) {
	AVCodecContext *c;
	AVStream *st;
	int retflag = TRUE;

	st = av_new_stream(oc, 1);
	if (!st) {
		fprintf(stderr, "Could not alloc stream\n");
		retflag = FALSE;
	}
	
	if (retflag != FALSE) {
		c = st->codec;
		c->codec_id = codec_id;
		c->codec_type = CODEC_TYPE_AUDIO;

		/* put sample parameters */
		//c->bit_rate = 64000;
		c->sample_rate = sample_rate;
		c->channels = channels;
	}
	
	if(retflag != FALSE)	
		return st;
	else
		return NULL;
}

int ff_decode_audio(const char* outfilename, const char *filename, 
					 const char *timepos_start, const char *timepos_stop, const int rangedecoding,
					 const char *cachefilename) {
	AVCodec *codec=NULL;
	AVPacket pkt;
	AVCodecContext *c=NULL, *ccx=NULL;
	AVFormatContext *fcx=NULL, *ofcx=NULL;
	AVOutputFormat *fmt=NULL;
	AVStream *audio_st=NULL;
	AVFormatParameters params;
	int got_audio, len, err=0, audio_index=-1, i=0, counter=0, j=0, bytecounter=0, channels=-1, pack_duration=0, packsize=0, codec_is_open=-1;
	long int file_size=0, header_size=0;
	char sample;
	int step = sizeof(sample)*2;
	char *maxsample=NULL, *minsample=NULL;
	int charsize = sizeof(char);
	float ratewindow=-1;
	unsigned int offset=0, maxbyte=0, sampledcounter=1;
	long long int seek_time_start=0, seek_time_stop=0, pack_pts=0;
	FILE *outfile=NULL, *cachefile=NULL;
	char *outbuf;
	unsigned char *packptr;
	int ret = TRUE, nobrk = TRUE;
	int endianess = isLittleEndian();
	outbuf = malloc(AVCODEC_MAX_AUDIO_FRAME_SIZE);

	av_register_all();

	// Open the input/output files
	err = av_open_input_file(&fcx, filename, NULL, 0, &params);
	if(err<0){
		printf("Can't open file: %s\n", filename);
		ret = FALSE;
	}
	
	if(cachefilename) {
		cachefile = fopen(cachefilename, "wb");
		if (!cachefile) {
			fprintf(stderr, "Could not open '%s'\n", cachefilename);
			ret = FALSE;
		}
	}

	if (ret != FALSE) {	
		// Find the stream info.
		av_find_stream_info(fcx);

		/* Find the first supported codec in the audio streams */
		for(i=0; i<fcx->nb_streams; i++){
			ccx=fcx->streams[i]->codec;
			if(ccx->codec_type==CODEC_TYPE_AUDIO) {
				/* Found an audio stream, check if codec is supported */
				codec = avcodec_find_decoder(ccx->codec_id);
				if(codec){
					/* codec is supported, proceed */
					audio_index = i;
					break;
				}
			}
		}
	}

	/* Find codec id */
	if(audio_index < 0) {
	 	printf("Audio stream with supported codec not found.\n");
	 	ret = FALSE;
	}
	else {
		c = avcodec_alloc_context();
		/* open it */
		if ((codec_is_open = avcodec_open(c, codec)) < 0) {
			fprintf(stderr, "could not open codec\n");
			ret = FALSE;
		}
		else {
			// See how many channels is the input file
			if(cachefilename) {
				channels = ccx->channels;
				ratewindow = (float)ccx->sample_rate / RESOLUTION;
				maxbyte = (sampledcounter) * channels * step * ratewindow;
				maxsample = malloc(channels);
				minsample = malloc(channels);
		
				// Initialize these arrays
				for (j=1;j<=channels;j++) {
					minsample[j-1] = 127;
					maxsample[j-1] = -128;
				}
				printf("Audio decoding...\n\n");
			}

			if (rangedecoding) {
				seek_time_start = parse_date(timepos_start,1);
				seek_time_stop = parse_date(timepos_stop,1);
				/* Check that timepos_start is smaller than timepos_stop */
				if (seek_time_start < seek_time_stop) {
					/* Check that timepos is inside duration */
					if (seek_time_start <= fcx->duration && seek_time_stop <= fcx->duration) {
						/* All ok */
						/*  Seek to the nearest keyframe before the seek_time we asked */
						/* Notice: always seek to stream 0 because if it is an avi 
						 * we need to seek by the video stream. not the audio */
						av_seek_frame(fcx, 0, av_rescale_q(seek_time_start, AV_TIME_BASE_Q, fcx->streams[0]->time_base), AVSEEK_FLAG_BACKWARD);
					}
					else {
						fprintf(stderr, "Seek start/stop time cannot be greater than input file's duration.\nExiting...\n");
						ret = FALSE;
					}
				}
				else {
					fprintf(stderr, "Seek start time is not smaller than seek stop time.\nExiting ...\n");
					ret = FALSE;
				}
			}
		}
	}

	/* Create wav headers */
	if (!cachefilename && ret != FALSE) {
		// allocate the output media context
		ofcx = av_alloc_format_context();

		if (!ofcx) {
			printf("Memory error\n");
			ret = FALSE;
		}

		// use wav as the output format of the file
		fmt = guess_format("wav", NULL, NULL);
		if (!fmt) {
			printf("Could not find suitable output format\n");
			ret = FALSE;
		}

		if(ret != FALSE) {
			ofcx->oformat = fmt;
			snprintf(ofcx->filename, sizeof(ofcx->filename), "%s", outfilename);

			// add the audio stream using the default format codec and initialize the codec
			if (fmt->audio_codec != CODEC_ID_NONE) {
				audio_st = ff_add_audio_stream(ofcx, fmt->audio_codec, ccx->sample_rate, ccx->channels);
			}

			// set the output parameters (must be done even if no parameters)
			if (av_set_parameters(ofcx, NULL) < 0) {
				printf("Invalid output format parameters\n");
				ret = FALSE;
			}
			else {
				// open the output file, if needed
				if (!(fmt->flags & AVFMT_NOFILE)) {
					if (url_fopen(&ofcx->pb, outfilename, URL_WRONLY) < 0) {
						fprintf(stderr, "Could not open '%s'\n", outfilename);
						ret = FALSE;
					}
				}

				// write the stream header, if any
				if (av_write_header(ofcx) < 0) {
					printf("Could not write header for output file\n") ;
					ret = FALSE;
				}
				else {
					// Clean up part 1
					// free the streams
					for(i = 0; i < ofcx->nb_streams; i++) {
						av_freep(&ofcx->streams[i]);
					}
				}
			}
		}

		// Clean up part 2
		if (!(fmt->flags & AVFMT_NOFILE)) {
			// close the output file
			url_fclose(&ofcx->pb);
		}

		// free the stream
		if(ofcx != NULL) av_free(ofcx);

		/* Create WAV headers end */
		if (ret != FALSE) {
			outfile = fopen(outfilename, "r+b");
			if (!outfile) {
				printf("Could not open '%s'\n", outfilename);
				ret = FALSE;
			}
			else {
				// truncate the last four 0 bytes of the empty header
				fseek(outfile, -4, SEEK_END);
				header_size = ftell(outfile);
				printf("Audio decoding...\n\n");
			}
		}
	}

	while(ret != FALSE && nobrk != FALSE) {
		// Read a frame/packet
		if(av_read_frame(fcx, &pkt) < 0 ) break;
		packsize = pkt.size;
		packptr = pkt.data;
		// Make sure this packet belongs to the stream we want
		if(pkt.stream_index==audio_index) {
			while (packsize > 0 && nobrk != FALSE) {
				// Rescale the times
				pack_pts = av_rescale_q(pkt.pts, fcx->streams[audio_index]->time_base, AV_TIME_BASE_Q);
				pack_duration = av_rescale_q(pkt.duration, fcx->streams[audio_index]->time_base, AV_TIME_BASE_Q);

				// Decode the paket
				len = avcodec_decode_audio(c, (short *)outbuf, &got_audio, packptr, packsize);

				if (len < 0) {
					fprintf(stderr, "Error while decoding\n");
					ret = FALSE;
				}

				packsize -= len;
				packptr += len;

				if ( got_audio > 0 ) {
					// If we got the time exactly, or we are already past the seek time,
					// this is the frame we want
					if (rangedecoding && pack_pts >= seek_time_start) {
						// if a frame has been decoded, output it
						if(!cachefilename) ff_audio_out(outbuf, 1, got_audio, outfile);

						// if the next frame gets past our stop time, we want to stop decoding
						if ( pack_pts + pack_duration > seek_time_stop ) {
							av_free_packet(&pkt);
							nobrk = FALSE;
							break;
						}
					}
					// If the next frame will be past our start seek time, this is the frame we want
					else if (rangedecoding && pack_pts + pack_duration > seek_time_start) {
						// if a frame has been decoded, output it
						if(!cachefilename) ff_audio_out(outbuf, 1, got_audio, outfile);

						// if the next frame gets past our stop time, we want to stop decoding
						if ( pack_pts + pack_duration > seek_time_stop ) {
							av_free_packet(&pkt);
							nobrk = FALSE;
							break;
						}
					}
					else if (!rangedecoding) {
						// if a frame has been decoded, output it
						if(!cachefilename) ff_audio_out(outbuf, 1, got_audio, outfile);
					}

					// Output cache
					if(cachefilename) {
						// bytecounter counts for each 'chunk' avcodec gives us, not the whole stream
						// bytecounter+offset is our current position in the whole stream
						bytecounter=0;
						// got_audio is the length of the 'chunk'
						while (bytecounter < got_audio) {
							// we check if the byte of the next sample is beyond the byte of the downsampling window
							if (bytecounter+offset + step*channels >= maxbyte) {
								for (j = 1; j <= channels ; j++) {
									// actually sample is not the whole sample, but the data for one channel each time
									sample = (char)outbuf[bytecounter+endianess];
									// min max averaging: only keep the highest and the lowest sample value
									if (maxsample[j-1] < sample) maxsample[j-1] = sample;
									if (minsample[j-1] > sample) minsample[j-1] = sample;
									fwrite(&maxsample[j-1], charsize, 1, cachefile);
									fwrite(&minsample[j-1], charsize, 1, cachefile);
									// reset min and max
									minsample[j-1] = 127;
									maxsample[j-1] = -128;
									// move 'chunk' byte pointer to current sample, next channel
									bytecounter += step;
								}
								// increase the counter showing how many samples we've averaged
								sampledcounter++;
								// move downsampling window pointer to the next window position
								maxbyte = (sampledcounter) * channels * step * ratewindow;
							}
							else {
								for (j = 1; j <= channels; j++) {
									sample = (char)outbuf[bytecounter+endianess];
									// min max averaging
									// if you want only downsampling, comment min max
									if (maxsample[j-1] < sample) maxsample[j-1] = sample;
									if (minsample[j-1] > sample) minsample[j-1] = sample;
									// move 'chunk' byte pointer to current sample, but next channel
									bytecounter += step;
								}
							}
						}
					}
					// Display the timestamp as progress (eye candy)
					ff_display_decoding_progress(pack_pts, counter);
					counter++;
					// add to the offset our current position
					offset += bytecounter;
				}
			}
		}
		av_free_packet(&pkt);
	}

	/* Write Trailer */
	if (!cachefilename && ret != FALSE ) {
		file_size = ftell(outfile);
		fseek(outfile, header_size, SEEK_SET);
		fput_le32(outfile, file_size - header_size - sizeof(file_size));
	}

	/* Clean up */
	printf("\n");
	if (!cachefilename) {
		if (outfile != NULL) fclose(outfile);
	}
	if (cachefilename) {
		if(maxsample != NULL) free(maxsample);
		if(minsample != NULL) free(minsample);
		if (cachefile != NULL) fclose(cachefile);
	}
	if (codec_is_open >= 0) avcodec_close(c);
	if (c != NULL )         free(c);
	if (outbuf != NULL)     av_free(outbuf);
	if (fcx != NULL)        av_close_input_file(fcx);

	return ret;
}

size_t ff_audio_out(char *ptr, size_t size, size_t nmemb, FILE *stream) {
	char swap;
	int i=0;

	if (!isLittleEndian()) {
		for (i=0; i<= ((size*nmemb)-1); i++) {
			swap = ptr[i]; // high
			ptr[i] = ptr[i+1]; // low
			ptr[++i] = swap ; // putting high back
		}
		
	}

	return fwrite(ptr, size, nmemb, stream);
}
