/*
 *   Copyright (c) 2006 by Thanos Kyritsis
 *
 *   This file is part of ffpocket
 *
 *   FFpocket is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   FFpocket is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with FFpocket; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


#include "ffdecode.h"


int ff_ppm_out(AVFrame *rgb, size_t cols, size_t rows, int pixsize, const char *filename) {
	int y = 0;
	FILE *file;

	// Open file
	file = fopen(filename, "wb");
	if(file == NULL) {
		fprintf(stderr, "\nCould not open '%s'\n", filename);
		return TRUE;
	}

	// Write header
	fprintf(file, "P6 %d %d 255\n", cols, rows);

	// Write pixel data
	for(y=0; y<rows; y++)
		fwrite(rgb->data[0]+y*rgb->linesize[0], 1, cols*3, file);

	// Close file
	fclose(file);
	return FALSE;
}

int ff_decode_frame(const char *input_filename, const char *timepos, const int onlyoneframe) {
	int err=0, ppmerr=0, i=0, counter=0, len=0, got_picture, video_index=-1, pack_duration=0, codec_is_open=-1;
	long long int seek_time=0, pack_pts=0;
	char imgfilename[16];
	int ret = TRUE;
	
	AVFormatContext *fcx=NULL;
	AVCodecContext *ccx=NULL;
	AVCodec *codec=NULL;

	AVFormatParameters params;
	AVPacket pkt;
	AVFrame *frame=avcodec_alloc_frame();
	AVFrame *pict=avcodec_alloc_frame();

	// Open the input file.
	err = av_open_input_file(&fcx, input_filename, NULL, 0, &params);
	if(err<0){
		printf("Can't open file: %s\n", input_filename);
		ret = FALSE;
	}

	if (ret != FALSE) {
		// Find the stream info.
		av_find_stream_info(fcx);

		// Find the first supported codec in the video streams.
		for(i=0; i<fcx->nb_streams; i++){
			ccx=fcx->streams[i]->codec;
			if(ccx->codec_type==CODEC_TYPE_VIDEO) {
				/* Found a video stream, check if codec is supported */
				codec = avcodec_find_decoder(ccx->codec_id);
				if(codec) {
					// codec is supported, proceed
					video_index=i;
					break;
				}
			}
		}
	}

	if (video_index < 0) {
		printf("Video stream with supported codec not found.\n");
		ret = FALSE;
	}
	else {
	// Open codec
		if((codec_is_open = avcodec_open(ccx, codec)) < 0) {
			fprintf(stderr, "could not open codec\n");
			ret = FALSE;
		}
		else {
			if (onlyoneframe) {
				/* Format seek time as necessary */
				seek_time = parse_date(timepos,1);

				if (seek_time > fcx->duration) {
					fprintf(stderr, "The seek time you requested is beyond the duration of the input file.\nExiting ...\n");
					ret = FALSE;
				}

				/*  Seek to the nearest keyframe before the seek_time we asked */
				av_seek_frame(fcx, video_index, av_rescale_q(seek_time, AV_TIME_BASE_Q, fcx->streams[video_index]->time_base), AVSEEK_FLAG_BACKWARD);
			}
		}
	}


	/* Decode proper */
	while(ret != FALSE) {
		// Read a frame/packet
		if(av_read_frame(fcx, &pkt) < 0 ) break;
		// Make sure this packet belongs to the stream we want
		if(pkt.stream_index==video_index){
			// Rescale the times
			pack_pts = av_rescale_q(pkt.pts, fcx->streams[video_index]->time_base, AV_TIME_BASE_Q);
			pack_duration = av_rescale_q(pkt.duration, fcx->streams[video_index]->time_base, AV_TIME_BASE_Q);
			// Decode this packet
			len = avcodec_decode_video(ccx, frame, &got_picture, pkt.data, pkt.size);
			if(len < 0) {
				printf("Error while decoding\n");
				ret = FALSE;
				av_free_packet(&pkt);
				break;
			}
			if (got_picture) {
				// If we got the time exactly, or we are already past the seek time, 
				// this is the frame we want
				if (onlyoneframe && pack_pts >= seek_time) {
					ff_display_decoding_progress(pack_pts, counter);
					av_free_packet(&pkt);
					break;
				}
				// If the next frame will be past our seek_time, this is the frame we want
				else if (onlyoneframe && pack_pts + pack_duration > seek_time) {
					ff_display_decoding_progress(pack_pts, counter);
					av_free_packet(&pkt);
					break;
				}
				else if (!onlyoneframe) {
					sprintf(imgfilename, "frame%06d.ppm", counter);
					if(counter==0)
						avpicture_alloc((AVPicture *)pict, PIX_FMT_RGB24, ccx->width, ccx->height);
					img_convert((AVPicture *)pict, PIX_FMT_RGB24, (AVPicture*) frame, ccx->pix_fmt, ccx->width, ccx->height);
					err = ff_ppm_out(pict, ccx->width, ccx->height, 3, imgfilename);
					if (ppmerr == TRUE) {
						ret = FALSE;
						av_free_packet(&pkt);
						break;
					}
					ff_display_decoding_progress(pack_pts, counter);
				}
				counter++;
			}
		}
	}

	if(onlyoneframe && ret != FALSE) {
		/* Allocate an AVPicture */
		avpicture_alloc((AVPicture *)pict, PIX_FMT_RGB24, ccx->width, ccx->height);
		img_convert((AVPicture *)pict, PIX_FMT_RGB24, (AVPicture*) frame, ccx->pix_fmt, ccx->width, ccx->height);
		/* save image to disk */
		err = ff_ppm_out(pict, ccx->width, ccx->height, 3, "frame.ppm");
		if (ppmerr == TRUE) {
			ret = FALSE;
		}
	}


	/* Clean up */
	printf("\n");
	if(ret != FALSE)       av_free(pict);
	if(frame != FALSE)     av_free(frame);
	if(codec_is_open >= 0) avcodec_close(ccx);
	if(fcx != FALSE)       av_close_input_file(fcx);

	return ret;
}

