/*
 *   Copyright (c) 2006 by Thanos Kyritsis
 *
 *   This file is part of ffpocket
 *
 *   FFpocket is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   FFpocket is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with FFpocket; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include "ffpocket.h"


void usage(const char *prog, const char *message) {
	printf("%s Version: %s\n", FFPOCKET_APPNAME, FFPOCKET_VERSION); 
	printf("Usage: %s [OPTIONS] -i FILE\n", FFPOCKET_APPNAME); 

	printf("\nOptions: \n");
	printf("  -i FILE                Set input filename to FILE\n");
	printf("  -o FILE                Set output filename to FILE\n");
	printf("  -c FILE                Set cache filename to FILE\n");
	printf("  --ss TIMESTAMP\t Set seek start time to timestamp\n\t\t\t (format is hh:mm:ss.xxx)\n");
	printf("  --st TIMESTAMP\t Set seek stop time to timestamp\n\t\t\t (format is hh:mm:ss.xxx)\n");
	printf("  --audio                Decode audio from input file to output file\n");
	printf("  --info                 Dump information about the given media file\n");
	printf("  -h, --help             This help information\n\n");

	printf("The default action is to decode frames, not audio.\n");
	printf("If timestamp is not given, then the whole input file will be decoded to frames.\n\n");

	printf("If audio decoding is specified but not both timestamps (start/stop),\n");
	printf("then the whole stream will be decoded.\n\n");

	printf("Examples: \n");
	printf("Decode a whole avi to frames: \n");
	printf("%s -i lala.avi \n", FFPOCKET_APPNAME);
	printf("Decode only one frame, the one located right on the timestamp given:\n");
	printf("%s -i lala.avi --ss hh:mm:ss.xxx \n", FFPOCKET_APPNAME);
	printf("Decode the whole audio stream of the avi: \n");
	printf("%s -i lala.avi -o koko.wav --audio \n", FFPOCKET_APPNAME);
	printf("Decode from --ss to --st: \n");
	printf("%s -i lala.avi -o koko.wav --audio --ss hh:mm:ss.xxx --st hh:mm:ss.xxx \n", FFPOCKET_APPNAME);
	printf("\n");
	

	if (message) {
		printf("%s\n\n", message);
	}
}

void file_info(const char *input_filename) {
	int err=0;
	AVFormatContext * fcx=NULL;
	AVFormatParameters params;

	av_register_all();

	// Open the input file.
	err = av_open_input_file(&fcx, input_filename, NULL, 0, &params);
	if(err<0){
		printf("Can't open file: %s\n", input_filename);
		exit(1);
	}

	// Find the stream info
	err = av_find_stream_info(fcx);
	
	// Give us information about the file and exit
	dump_format(fcx, 0, input_filename, FALSE);

	av_close_input_file(fcx);
}

int main(int argc, char * argv[]) {
	int opt, showinfo=0, decodeaudio=0, decodeframe=1;
	char *timepos_start=NULL, *timepos_stop=NULL, *outfile=NULL, *infile=NULL, *cachefile=NULL;
	static struct option long_options[] = {
		{"help", 0, 0, 'h'},
		{"info", 0, 0, 'n'},
		{"ss", 1, 0, 's'},
		{"st", 1, 0, 't'},
		{"audio", 0, 0, 'a'},
		{0,0,0,0}
	};
	 
	while((opt=getopt_long(argc, argv, "c:o:i:h", long_options, NULL))!=-1){
		switch(opt) {
			case 'i':
				infile = optarg;
				break;
			case 'o':
				outfile = optarg;
				break;
			case 's':
				timepos_start = optarg;
				break;
			case 't':
				timepos_stop = optarg;
				break;
			case 'n':
				showinfo = 1;
				decodeframe = 0;
				break;
			case 'a':
				decodeaudio = 1;
				decodeframe = 0;
				break;
			case 'c':
				cachefile = optarg;
				break;
			case 'h':
			default:
				usage(FFPOCKET_APPNAME, NULL);
				exit(1);
				break;
		}
	}

	if (argc < 2 ) {
		usage(FFPOCKET_APPNAME, "You must specify some options.");
		return 1;
	}

	if (showinfo) {
		if (infile)
			file_info(infile);
		else {
			usage(FFPOCKET_APPNAME, "You must specify an input file.");
			return 1;
		}
	}

	if (infile && outfile) {
		if (!strcmp(infile,outfile)) {
			usage(FFPOCKET_APPNAME, "You provided the same input and output file!!");
			return 1;
		}
	}

	if (infile && cachefile) {
		if (!strcmp(infile,cachefile)) {
			usage(FFPOCKET_APPNAME, "You provided the same input and cache file!!");
			return 1;
		}
	}

	if (decodeframe){
		if(infile && timepos_start) {
			av_register_all();
			ff_decode_frame(infile, timepos_start, TRUE);
			return 0;
		}
		else if(infile && !timepos_start) {
			av_register_all();
			ff_decode_frame(infile, timepos_start, FALSE);
			return 0;
		}
		else {
			usage(FFPOCKET_APPNAME, "You must specify an input file and optionally a seek time position.");
			return 1;
		}
	}

	if (decodeaudio){
		if(infile && (outfile||cachefile) && !(timepos_start && timepos_stop)) {
			ff_decode_audio(outfile, infile, timepos_start, timepos_stop, FALSE, cachefile);
			return 0;
		}
		else if (infile && (outfile||cachefile)&& timepos_start && timepos_stop) {
			ff_decode_audio(outfile, infile, timepos_start, timepos_stop, TRUE, cachefile);
			return 0;
		}
		else {
			usage(FFPOCKET_APPNAME, "You must specify BOTH an input AND an output (OR cache) file.");
			return 1;
		}
	}

	return 0;
}

